
const prettify = require('object-inspect');


type Result<T> = [T, Array<Fault>];
type Fault = { path: Array<string>, issue: string };

function mkfault(issue: string): Fault {
  return { path: [], issue };
}

function mkfaults(...issues: Array<string>): Array<Fault> {
  return issues.map(mkfault);
}


class Coerce<T> {

  run: (dirty: unknown) => Result<T>;

  get fallback(): T {
    return this.run(undefined)[0];
  }

  constructor(run: Coerce<T>['run']) {
    this.run = run;
  }

  go(dirty: unknown): T {
    const [clean, faults] = this.run(dirty);

    if (faults.length > 0) {
      let warning = '[kempt warning] Invalid value encountered'
      warning += `\n\nReceived val = ${prettify(dirty)}`;
      warning += '\n\nIssues:'
      for (const fault of faults)
        warning += `\n  - val${joinPath(fault.path)} ${fault.issue}`;
      warning += `\n\nValue patched to: ${prettify(clean)}`;
      console.warn(warning);
    }
    
    return clean;

    function joinPath(path: Array<string>) {
      return path.map(key => '1234567890'.includes(key[0]) ? `[${key}]` : `.${key}`).join('');
    }
  }
  
}


type Critic = (dirty: unknown) => string | false;
type Fix<T> = { test: (dirty: unknown) => boolean, run: (dirty: unknown) => T };

function k<T>(test: Critic, fallback: T, fixes: Array<Fix<T>> = []): Coerce<T> {

  const fault = test(fallback);
  if (fault) throw Error('Provided fallback is invalid.');
  
  return new Coerce((dirty: unknown): Result<T> => {

    const faults: Array<Fault> = [];
    
    const issue = test(dirty);
    if (!issue) return [dirty as T, []];
    faults.push(mkfault(issue));
    
    for (let idx = 0; idx < fixes.length; idx++) {
      const fix = fixes[idx];

      let passed;
      try {
        passed = fix.test(dirty);
      } catch (ex) {
        faults.push(mkfault(`Fix #${idx} failed (test function crashed).`));
        continue;
      }

      if (!passed)
        continue;

      let clean;
      try {
        clean = fix.run(dirty);
      } catch (ex) {
        faults.push(mkfault(`Fix #${idx} failed (coerce function crashed).`));
        continue;
      }

      const cleanFault = test(clean);
      if (typeof cleanFault === 'string') {
        faults.push(mkfault(`Fix #${idx} failed (produced indirtyid dirtyue).`));
        continue;
      }

      return [clean, faults];
    }

    return [fallback, faults];
    
  });
}



const nully: Critic = dirty =>
  dirty !== null && dirty !== undefined && "isn't null or undefined";

// includes NaN or infinities
const float: Critic = dirty =>
  typeof dirty !== "number" && "isn't a number";

const number: Critic = dirty => 
  float(dirty)
  || Number.isNaN(dirty) && 'is NaN'
  || !Number.isFinite(dirty) && "isn't finite";

const integer: Critic = dirty =>
  number(dirty)
  || !Number.isInteger(dirty) && "isn't an integer";

const bigint: Critic = dirty =>
  !(dirty instanceof BigInt) && "isn't a bigint";

const string: Critic = dirty =>
  typeof dirty !== 'string' && "isn't a string";

const boolean: Critic = dirty =>
  typeof dirty !== 'boolean' && "isn't a boolean";
  

function Object_map<V, R>(
    obj: Record<string, V>,
    mapper: (key: string, val: V) => [string, R],
  ): Record<string, R>
{
  return Object.fromEntries(Object.entries(obj).map(([key, val]) => mapper(key, val)));
}

function bump(faults: Array<Fault>, key: any) {
  return faults.map(({ path, issue }) => ({ path: ['' + key, ...path], issue }));
}

function object<Obj extends Record<keyof any, unknown>>(
    shape: { [K in keyof Obj & string]: Coerce<Obj[K]> }
  ): Coerce<Pick<Obj, string>>
{

  if (shape === null || typeof shape !== 'object')
    throw Error('Expected an object');

  for (const coerce of Object.values(shape))
    if (!(coerce instanceof Coerce))
      throw Error('Expected a Coerce');

  return new Coerce((dirty: unknown): Result<Pick<Obj, string>> => {
    if (dirty === null || typeof dirty !== 'object') {
      const fallback = Object_map(shape, (key, coerce) => [key, coerce.fallback]);
      return [fallback, mkfaults("isn't an object")];
    }
    const dirtyr = dirty as Record<string, unknown>;

    const faults = [];

    if (![Object.prototype, null].includes(Object.getPrototypeOf(dirty)))
      faults.push(mkfault('has a prototype other than Object.prototype or null'));

    faults.push(...Object.entries(shape).flatMap(([key, coerce]) => bump(coerce.run(dirtyr[key])[1], key)));
    faults.push(...Object.keys(dirtyr).filter(key => !(key in shape)).flatMap(key => bump(mkfaults('is superfluous'), key)));
    const result = Object_map(shape, (key, coerce) => [key, coerce.run(dirtyr[key])[0]]);
    return [result, faults];
  });
  
}

function checkLength(expected: number, actual: number): [] | [Fault] {
  if (actual === expected - 1)
    return [{ path: [`${actual}`], issue: 'is missing' }];
  else if (actual < expected)
    return [{ path: [`${actual}..${expected - 1}`], issue: 'are missing' }];
  else if (actual === expected + 1)
    return [{ path: [`${expected}`], issue: 'is out of bounds' }];
  else if (actual > expected)
    return [{ path: [`${expected}..${actual - 1}`], issue: 'are out of bounds' }];
  else
    return [];
}

function tuple<Tup extends Array<unknown>>(...shape: { [K in keyof Tup]: Coerce<Tup[K]> }): Coerce<Tup> {

  if (!(shape instanceof Array))
    throw Error('Expected Array');

  for (const item of shape)
    if (!(item instanceof Coerce))
      throw Error('Expected Coerce');

  return new Coerce((dirty: unknown): Result<Tup> => {
    if (!(dirty instanceof Array)) {
      const fallback = shape.map(coerce => coerce.fallback) as Tup;
      return [fallback, mkfaults("isn't an array")];
    }

    const faults = [];
    faults.push(...checkLength(shape.length, dirty.length));
    faults.push(...shape
      .slice(Math.min(shape.length, dirty.length))  // Don't repeat out-of-bounds faults
      .flatMap((coerce, idx) => bump(coerce.run(dirty[idx])[1], idx)));


    const result = shape.map((coerce, idx) => coerce.run(dirty[idx])[0]) as Tup;
    return [result, faults];
  });
  
}

function vector<T>(coerce: Coerce<T>, length: number): Coerce<Array<T>> {

  if (!(coerce instanceof Coerce))
    throw Error('Expected Coerce');

  return new Coerce((dirty: unknown): Result<Array<T>> => {
    if (!(dirty instanceof Array)) {
      const fallback = Array(length).fill(null).map(() => coerce.fallback);
      return [fallback, mkfaults("isn't an array")];
    }

    const faults = dirty.flatMap((elem, idx) => bump(coerce.run(dirty[idx])[1], idx));

    faults.push(...checkLength(length, dirty.length));
     
    const result = dirty.map(elem => coerce.run(elem)[0]);
    return [result, faults];
  });
  
}

function list<T>(coerce: Coerce<T>): Coerce<Array<T>> {

  if (!(coerce instanceof Coerce))
    throw Error('Expected Coerce');

  return new Coerce((dirty: unknown): Result<Array<T>> => {
    if (!(dirty instanceof Array)) {
      return [[], mkfaults("isn't an array")];
    }

    const faults = dirty.flatMap((elem, idx) => bump(coerce.run(dirty[idx])[1], idx));
    const result = dirty.map((elem, idx) => coerce.run(dirty[idx])[0]);
    return [result, faults];
  });
  
}






const coerce = object({
  latex: k(string, ''),
  inline: k(boolean, false),
});
const result = coerce.go({ x: Number.EPSILON, y: 3 });
